@@ -13,7 +13,7 @@ struct xbuf {
 
 struct bfd {
 	int fd;
-	int mode;
+	bool writable;
 	struct xbuf b;
 };
 
@@ -27,7 +27,8 @@ static inline void bfd_setraw(struct bfd *b)
 	b->b.mem = NULL;
 }
 
-int bfdopen(struct bfd *f, int mode);
+int bfdopenr(struct bfd *f);
+int bfdopenw(struct bfd *f);
 void bclose(struct bfd *f);
 char *breadline(struct bfd *f);
 int bwrite(struct bfd *f, const void *buf, int sz);
