@@ -13,8 +13,8 @@
 struct fdinfo_entry;
 struct stat;
 
-extern int fd_id_generate(pid_t pid, FdinfoEntry *fe, struct stat *st);
-extern int fd_id_generate_special(struct stat *, u32 *id);
-extern void fd_id_show_tree(void);
+struct fd_parms;
+extern int fd_id_generate(pid_t pid, FdinfoEntry *fe, struct fd_parms *p);
+extern int fd_id_generate_special(struct fd_parms *p, u32 *id);
 
 #endif /* __CR_FILE_IDS_H__ */
