@@ -53,8 +53,8 @@ struct page_read {
 	void (*close)(struct page_read *);
 
 	/* Private data of reader */
-	int fd;
-	int fd_pg;
+	struct cr_img *pmi;
+	struct cr_img *pi;
 
 	PagemapEntry *pe;		/* current pagemap we are on */
 	struct page_read *parent;	/* parent pagemap (if ->in_parent
@@ -63,9 +63,28 @@ struct page_read {
 					   read_pagemap_page */
 	unsigned long cvaddr;		/* vaddr we are on */
 
+	struct iovec bunch;		/* record consequent neighbour
+					   iovecs to punch together */
 	unsigned id; /* for logging */
 };
 
-extern int open_page_read(int pid, struct page_read *);
+#define PR_SHMEM	0x1
+#define PR_TASK		0x2
 
+#define PR_TYPE_MASK	0x3
+#define PR_MOD		0x4	/* Will need to modify */
+
+/*
+ * -1 -- error
+ *  0 -- no images
+ *  1 -- opened
+ */
+extern int open_page_read(int pid, struct page_read *, int pr_flags);
+extern int open_page_read_at(int dfd, int pid, struct page_read *pr, int pr_flags);
+extern void pagemap2iovec(PagemapEntry *pe, struct iovec *iov);
+extern void iovec2pagemap(struct iovec *iov, PagemapEntry *pe);
+extern int seek_pagemap_page(struct page_read *pr, unsigned long vaddr, bool warn);
+
+extern int dedup_one_iovec(struct page_read *pr, struct iovec *iov);
+extern int punch_hole(struct page_read *pr, unsigned long off, unsigned long len, bool cleanup);
 #endif /* __CR_PAGE_READ_H__ */
