@@ -27,8 +27,7 @@ function launch_test {
 function stop_test {
 	wtime=1
 	cd ../zdtm/live/static/
-	make maps04.out
-	while [ ! -f maps04.out ]; do echo "Wait for maps04.out"; sleep $wtime; ((wtime++)); done
+	make maps04.stop
 	cat maps04.out | fgrep PASS || fail "Test failed"
 	echo "OK"
 }
@@ -58,7 +57,7 @@ ${CRIU} pre-dump -D "$IMGDIR/dump-2/pre/" -o dump.log -t ${PID} -v4 || fail "Fai
 
 echo "Plain dump"
 mkdir "$IMGDIR/dump-2/plain/"
-${CRIU} dump -D "$IMGDIR/dump-2/plain/" -o dump.log -t ${PID} -v4 --prev-images-dir=../pre/ || fail "Fail to dump"
+${CRIU} dump -D "$IMGDIR/dump-2/plain/" -o dump.log -t ${PID} -v4 --prev-images-dir=../pre/ --track-mem || fail "Fail to dump"
 
 sleep 1
 echo "Restore"
